#include <iostream>
#include <iomanip>
#include <fstream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "rof-event.h"
#include "cell-event.h"

#ifdef _OPENMP
#include <omp.h>        // OpenMP
#endif

using namespace std;

//֐̐錾difinition of the function
int makeBasinList(int row, int col, class Cell** object, int*& upperGridNum);
void read2DValue(ROF** fROF, int Ndays, int ROFnum_row, int ROFnum_col);

int main()
{
	cout << "Reading parameters for runoff simulation." << endl;
	ifstream finParam("./param-event.dat");
	if(!finParam){
		cout << "param-event.dat cannot open." << endl;
	}

	// vZ̈ simulation range
	cout << "1) Reading parameters for study area." << endl;

	int wCol,wRow;
	float wSWLON, wSWLAT;

	finParam >> wCol; 
	finParam >> wRow; 
	cout << "  Number of col : " << wCol << "  Number of row : " << wRow << endl;

	finParam >> wSWLON; 
	finParam >> wSWLAT;
	cout << " Sorth West corner Longitude (degree) : " << wSWLON 
		 << " Latitude (deg) : " << wSWLAT << endl;

	const int Col = wCol;     // column number of study area
	const int Row = wRow;     // row number of study area
	const float SWLON = wSWLON; // South west corner of longitude (degree)
	const float SWLAT = wSWLAT; // South west corner of latiitude (degree)

	// ̓f[^̈ input data range
	cout << "2) Reading parameters for input data." << endl;

	int wROFnum_col, wROFnum_row;
	double wDelta_x, wDelta_y, wROF_WLON, wROF_SLAT;
	finParam >> wROFnum_col;
	finParam >> wROFnum_row;
	cout << "  Number of col : " << wROFnum_col 
		 << "  Number of row : " << wROFnum_row << endl;

	finParam >> wDelta_x;
	finParam >> wDelta_y;
	cout << "  Input data grid size of x-direction : " << wDelta_x << endl;
	cout << "  Input data grid size of y-direction : " << wDelta_y << endl;

	finParam >> wROF_WLON; 
	cout << "  Longitude of the center of the Southwest corner of input data (degree) : " << wROF_WLON << endl;
	finParam >> wROF_SLAT; 
	cout << "  Latitude of the center of the Southwest corner of input data (degree) : " << wROF_SLAT << endl;
													
	const int ROFnum_col = wROFnum_col;
	const int ROFnum_row = wROFnum_row; 
	const float Delta_x = wDelta_x;
	const float Delta_y = wDelta_y;
	const float ROF_WLON = wROF_WLON;
	const float ROF_SLAT = wROF_SLAT;

	// vZ simulation term
	cout << "3) Reading parameters for simuration duration." << endl;
	int wNdays;
	finParam >> wNdays;
	cout << "  Simulation duration (days) : " << wNdays << endl;

	const int Ndays = wNdays; // Simulation duration (days)

	//  finite difference condition
	cout << "4) Reading parameters for finite difference condition." << endl;
	int wTime_Step, wNspace; double wTheta;
	finParam >> wTime_Step; cout << "  Finite difference time step (sec) : " << wTime_Step << endl;
	finParam >> wNspace; cout << "  Finite difference space step : " << wNspace << endl;
	finParam >> wTheta; cout << "  Finite difference weight : " << wTheta << endl;

	const int Time_Step = wTime_Step; // vZ̎ԊԊu [sec] similation time step [sec] */
	const int Nspace = wNspace; /* vZ̋ԕ number of spatial divisions */
	const double Theta = wTheta; 	/* d݌W (-) 0.5 < theta <= 1.0 */
	const int Ntime = 3600/Time_Step;  // ̓f[^Ԋu(1)ł̌vZ

	// model parameter setting for runoff simulation vZ̃p^
	cout << "5) Reading parameters for kinematic wave model." << endl;
	double wMan_channel, wMan_slope, wKa, wDa, wDm, wBeta; int wThreshold_basin;

	finParam >> wMan_slope; cout << "  Manning roughmess coefficient for slope [m-s unit] : " << wMan_slope << endl;
	finParam >> wKa; cout << "  Hydraulic conductivity [m-s] : " << wKa << endl;
	// finParam >> wD; cout << "  Toal soil layer [m] : " << wD << endl;
	finParam >> wDa; cout << "  Depth of capillary and non-capillary soil layers [m] : " << wDa << endl;
	finParam >> wDm; cout << "  Depth of capillary soil layers [m] : " << wDm << endl;
	finParam >> wBeta; cout << "  Exponent constant of unsaturated hydraulic conductivity : " << wBeta << endl;
	finParam >> wMan_channel; cout << "  Manning roughmess coefficient for channel [m-s unit] : " << wMan_channel << endl;

	finParam >> wThreshold_basin; cout << "  Area to classify slope and channel : " << wThreshold_basin << endl;
	const int Threshold_basin = wThreshold_basin; // WObh臒libasin > threshold_basin ݂Obh͓ɂȂj

	/* initial condi */
	cout << "6) Reading itinial discharge for kinematic wave model." << endl;
	double winitialROF;
	finParam >> winitialROF; cout << "  Initial runoff height [mm/hr] : " << winitialROF << endl;
	const double intialROF = winitialROF; /* initial runoff height [mm/hr] */

	finParam.close();
	//--------------------- end parameter reading ------------------------------------------------


	//ϐ̒` dif-variables
	int iday, ihour, dt;
	int n, k;
	bool executeOrPass = false;
	//errno_t error;
	
	FILE *fCorrectedMed;
	float med;

	FILE *fDown;
	FILE *fBasin;
	FILE *fRiverNumber;

	float ROFlat;
	float ROFlon;
	char DischargeFileName[256];
	int* upperGridNum;
	int numOfBasin;

	// reading watershed numbers for simulation from riverNumList.dat vZ闬̔ԍ̐ݒ
	ifstream fin1("./topoData/riverNumList.dat");
	if(!fin1){
		cout << "riverNumList.dat cannot open.\n";
		return 1;
	}

	int size_riverNumber_cal=0;
	string ss;

	while (getline(fin1, ss)) { // count number of rivers.  ΏۂƂ͐쐔𒲂ׂ
		size_riverNumber_cal++; 
	}
	fin1.close();
	cout << "Number of rivers for simulation : " << size_riverNumber_cal << endl << endl;

	int* riverNumber_cal;
	riverNumber_cal = new int[size_riverNumber_cal];

	ifstream fin2("./topoData/riverNumList.dat");
	if(!fin2){
		cout << "riverNumList.txt cannot open.\n";
		return 1;
	}
	int iriver, icol, irow, ilat, ilon, iarea;
	double dlon, dlat;
	for(int i=0; i<size_riverNumber_cal; i++){ // setting river number for simulation
		fin2 >> iriver >> iarea >> icol >> irow >> ilon >> ilat >> dlon >> dlat;
		riverNumber_cal[i]=iriver;
	}
	fin2.close();

	// Memory assignment for grid cell of simulation
	class Cell** cellObject; 
	cellObject = new class Cell*[Row];
	for(int i=0; i<Row; i++){
		cellObject[i] = new class Cell[Col];
	}

	// Memory assignment for data cell of input data
	class ROF** rofObject;
	rofObject = new class ROF*[ROFnum_row];
	for(int i=0; i<ROFnum_row; i++){
		rofObject[i] = new class ROF[ROFnum_col];
	}

	/*--- ̓f[^IuWFNgɏ ---*/
	//eIuWFNgɈܓxEox
	for(int i=0; i<ROFnum_row; i++){
		//ROFlat = ROF_SLAT + 0.5*Delta_y + i*Delta_y ;  // set latitude
		ROFlat = ROF_SLAT + i*Delta_y ;  // set latitude. Runoff grid is the center.
		for(int j=0; j<ROFnum_col; j++){
			//ROFlon = ROF_WLON +  0.5*Delta_x + j*Delta_x; // set longitude 
			ROFlon = ROF_WLON +  j*Delta_x; // set longitude  Runoff grid is the center.
			rofObject[i][j].setInformation(&ROFlat, &ROFlon, j, i, Ndays);
		}
	}
	//cout << "ROF data setting end." << endl;

	/*--- [͓ԁEoϐɒl set the sream informations ---*/

	/* latitude  longitude  longitudinalLength  med */
	//if(error = fopen_s(&fCorrectedMed, "./topoData/modDem.bin", "rb") != NULL) {
	if ((fCorrectedMed = fopen("./topoData/modDem.bin", "rb")) == NULL) {
		cout << "modDem.bin cannot open." << endl;
		exit(1);
	}

	for(int i=Row-1; i>=0; i--) { // DEM data arranges from south to north
		for(int j=0; j<Col; j++) {
			fread(&med, sizeof(float), 1, fCorrectedMed);
			cellObject[i][j].setElevation(i, j, Row, Col, SWLAT, SWLON, med);
			cellObject[i][j].setNULL();
			//cout << "row=" << i << " col=" << j << endl;
		}
	}
	fclose(fCorrectedMed);

	/* down  basin downStreamGrid riverNumber */
	//if(error = fopen_s(&fDown, "./topoData/flowDir.bin", "rb") != NULL) {
	if ((fDown = fopen("./topoData/flowDir.bin", "rb")) == NULL) {
		cout << "flowDir.bin cannot open." << endl;
		exit(1);
	}

	//if(error = fopen_s(&fBasin, "./topoData/flowAcc.bin", "rb") != NULL) {
	if ((fBasin = fopen("./topoData/flowAcc.bin", "rb")) == NULL) {
		cout << "flowAcc.bin cannot open." << endl;
		exit(1);
	}

	//if(error = fopen_s(&fRiverNumber, "./topoData/riverNum.bin", "rb") != NULL) {
	if ((fRiverNumber = fopen("./topoData/riverNum.bin", "rb")) == NULL) {
		cout << "riverNum.bin cannot open." << endl;
		exit(1);
	}

	for(int i=Row-1; i>=0; i--) { // Topo data arranges from south to north
		for(int j=0; j<Col; j++) {

			fread(&cellObject[i][j].down, sizeof(int), 1, fDown);
			fread(&cellObject[i][j].basin, sizeof(int), 1, fBasin);
			fread(&cellObject[i][j].riverNumber, sizeof(int), 1, fRiverNumber);

			if(cellObject[i][j].riverNumber < 0) {
				continue;
			}
			cellObject[i][j].setDownStreamGrid(i, j, cellObject);
		}
	}

	fclose(fDown);
	fclose(fBasin);
	fclose(fRiverNumber);

	/* upStreamGrid riverNumber */
	for(int i=0; i<Row; i++){
		for(int j=0; j<Col; j++)	{
			if(cellObject[i][j].riverNumber < 0 || cellObject[i][j].isThisGridEnd==true || cellObject[i][j].down==0) {
				continue;
			}
			//cout << i << ", " << j << endl;
			cellObject[i][j].setUpStreamGrid();
		}
	}

	/* numberOfUpStreamGrid */
	for(int i=0; i<Row; i++){
		for(int j=0; j<Col; j++){
			if(cellObject[i][j].riverNumber < 0 || cellObject[i][j].isThisGridEnd==true) {
				continue;
			}
			cellObject[i][j].countNumberOfUpStreamGrid();
		}
	}

	/* ROFCell ԗʂ p^ set parameters */
	for(int i=0; i<Row; i++){
		for(int j=0; j<Col; j++){
			
			if(cellObject[i][j].riverNumber < 0 || cellObject[i][j].isThisGridEnd == true) {
				continue;
			}

			cellObject[i][j].setROFCell(rofObject, ROFnum_row, ROFnum_col);
			cellObject[i][j].Set_parameter(wMan_channel, wMan_slope, wKa, wDa, wDm, wBeta, Threshold_basin, Nspace, Time_Step, Theta, Ntime, Ndays);
			cellObject[i][j].initialize(intialROF);

			cellObject[i][j].gridKind = 1;  // ΖʃObh
			if (cellObject[i][j].basin > Threshold_basin) {
				cellObject[i][j].gridKind = 2;  // ͐Obh
			}
		}
	}	

	//--- make basinList for setting runoff simulation order ---
	numOfBasin = makeBasinList(Row, Col, cellObject, upperGridNum);
	cout << "*** Number of different kind of cell accumulattion number : " << numOfBasin << endl;

	// --- ovZ runoff simulation ---
	cout << "Total simulation time is " << Ndays << " days." << endl << endl; 
	
	// Read 2-dimensional input data
	read2DValue(rofObject, Ndays, ROFnum_row, ROFnum_col);


	//=========Dam setting=====================================================================
	// observed dam outflow setting for dam 1
	class damCell damObject1;
	int damCol = 55; // from West starting from 1.
	int damRow =  2; // from South starting form 1.
	int damID = 99; // Dam ID number
	int damSimulationKind = 1; // 0: use observed dam outflow, 1: use dam model
	//damObject1.setDamGrid(damRow, damCol, Row, Col, damID, cellObject, damSimulationKind);
	//damObject1.setDamInitialState(0.0); // set initial dam storage ̓_fɉčςKvB
	//damObject1.setDamParameter(0.5); // set parameter ̓_fɉčςKvB

	// Reading monthly observed outflow for Dam1. Added on May 20, 2013. 
	// damObject1.readDamOutflow(iday, dt); ѕf[^pꍇɁA_f[^ǂݍ
	//=========================================================================================

	cout << "Runoff simulation starts." << endl; 
	for(int ib=0; ib<numOfBasin; ib++){ // ʐς̏̂珇ɌvZB
		cout << "  " <<  setprecision(3) << ((double)ib+1.0)/((double)numOfBasin)*100 << "% ends.\n";

		#pragma omp parallel for private(iday, ihour, dt, n, executeOrPass, k)
		for(int m=0; m<Row; m++){ // latitude direction
			for(n=0; n<Col; n++){ // longitude direction
				if(cellObject[m][n].isThisGridEnd==true) {
					continue;
				}
				// w肵ɑObĥ݌vZB
				// Only catchments specified in 'riverNumberList.txt' is calculated.
				executeOrPass = false;

				for(k=0; k<size_riverNumber_cal; k++){
					if(cellObject[m][n].riverNumber == riverNumber_cal[k]){
						executeOrPass = true;
						break;
					}
				}
				if(executeOrPass == false) {
					continue;
				}
				if(cellObject[m][n].basin == upperGridNum[ib]) {
					// runoff simulation
					for(iday=1; iday <= Ndays; iday++){ // day loop
						for(ihour=1; ihour<=24; ihour++){ // hour loop
							cellObject[m][n].getLateralRunoff(iday, ihour);
							for(dt=0; dt<Ntime; dt++){ // simulation time step roop 

								//if(cellObject[m][n].basin > Threshold_basin && cellObject[m][n].gridKind == 1){//͓Obh̏ꍇ
								//if(cellObject[m][n].basin > Threshold_basin) {//͓Obh̏ꍇ

								if(cellObject[m][n].gridKind == 2) {// ͓Obh̏ꍇ
									cellObject[m][n].CalculateRiver(iday, ihour, dt);
								}
								else if (cellObject[m][n].gridKind == 1) { //ΖʃObh̏ꍇ
									cellObject[m][n].CalculateBasin(iday, ihour, dt);
								}
								//else if (cellObject[m][n].gridKind == 99) { // _Obh̏ꍇ
								//	//cout << "Dam grid" << " row = " << m << " col = " << n << endl;
								//	damObject1.calculateDamOutflow(iday, ihour, dt);
								//}
							}
						}
					}
				}
			}
		} 			
	} // runoff simulation loop end. ovZIB

	cout << endl; 
	cout << "Runoff simulation ends." << endl; 

	// writing simulated data to files
	ofstream FileDaily; // Daily flow
	//sprintf_s(DischargeFileName, 256, "./output/dischargeDaily.bin");
	sprintf(DischargeFileName, "./output/dischargeDaily.bin");
	cout << "Write " << DischargeFileName << endl;
	FileDaily.open(DischargeFileName, ios::out | ios::binary);
	
	ofstream FileHourlyMax; // Maximum hourly data in a day
	//sprintf_s(DischargeFileName, 256, "./output/dischargeHourlyMaxInDay.bin");
	sprintf(DischargeFileName, "./output/dischargeHourlyMaxInDay.bin");
	cout << "Write " << DischargeFileName << endl;
	FileHourlyMax.open(DischargeFileName, ios::out | ios::binary);

	ofstream FileHourly; // Hourly data
	//sprintf_s(DischargeFileName, 256, "./output/dischargeHourly.bin");
	sprintf(DischargeFileName, "./output/dischargeHourly.bin");
	cout << "Write " << DischargeFileName << endl;
	FileHourly.open(DischargeFileName, ios::out | ios::binary);

	/*--- for debugging --------------------------------------------*/
	//ofstream FileTest; // Point hourly data
	//sprintf_s(DischargeFileName, 256, "./output/dischargeTest.dat");
	////sprintf(DischargeFileName, "./output/dischargeTest.dat");
	//cout << "Write " << DischargeFileName << endl;
	//FileTest.open(DischargeFileName);
	/*--- for debugging--------------------------------------------- */

	float Nodata=-9999;
	for(iday=0; iday < Ndays; iday++){
		float dailyFlow; // m3/hr
		float hourlyFlowMax; // m3/hr
		float hourlyFlow; // m3/hr
			
		for(int m=Row-1; m >=0; m--){ // latitude direction from South for Grads format
			for(n=0; n<Col; n++){     // longitude direction from West
				if(cellObject[m][n].isThisGridEnd==true){
					FileDaily.write((char*)&Nodata, sizeof(float));
					FileHourlyMax.write((char*)&Nodata, sizeof(float));
					continue;
				}
				// w肵ɑȂObhɂ false ݒ肷B 
				// 'false' is set to the grids of outside study area.
				executeOrPass = false;
				for(k=0; k<size_riverNumber_cal; k++){
					if(cellObject[m][n].riverNumber == riverNumber_cal[k]){
						executeOrPass = true;
						break;
					}
				}
				if(executeOrPass == false){ // ΏۊÕObhł Nodata Ń[vi߂B
					FileDaily.write((char*)&Nodata, sizeof(float));
				    FileHourlyMax.write((char*)&Nodata, sizeof(float));
					continue;
				}
					
				dailyFlow = (float) cellObject[m][n].GetDailyFlow(iday, Ntime);
				FileDaily.write((char*)&dailyFlow, sizeof(float));

				hourlyFlowMax = (float) cellObject[m][n].GetHourlyFlowMax(iday, Ntime);
				FileHourlyMax.write((char*)&hourlyFlowMax, sizeof(float));
			}
		}

		for (ihour=0; ihour < 24; ihour++) {
			for(int m=Row-1; m >=0; m--){ // latitude direction from South for Grads format
				for(n=0; n<Col; n++){     // longitude direction from West
					if(cellObject[m][n].isThisGridEnd==true){
						FileHourly.write((char*)&Nodata, sizeof(float));
						continue;
					}
					// w肵ɑȂObhɂ false ݒ肷B 
					// 'false' is set to the grids of outside study area.
					executeOrPass = false;
					for(k=0; k<size_riverNumber_cal; k++){
						if(cellObject[m][n].riverNumber == riverNumber_cal[k]){
							executeOrPass = true;
							break;
						}
					}
					if(executeOrPass == false){ // ΏۊÕObhł Nodata Ń[vi߂B
						FileHourly.write((char*)&Nodata, sizeof(float));
						continue;
					}

					hourlyFlow = (float) cellObject[m][n].GetHourlyFlow(iday, ihour, Ntime);
					FileHourly.write((char*)&hourlyFlow, sizeof(float));
				}
			}
		}
		/*--- File output for debugging by YT -------------------------------------------------*/
		//int outCol = 59; // begin from 1 from east
		//int outRow = 22; // begin from 1 from south
		//outRow = Row -outRow + 1; // change from north
		//for (ihour=0; ihour < 24; ihour++) {
		//	hourlyFlow = (float) cellObject[outRow-1][outCol-1].GetHourlyFlow(iday, ihour, Ntime);
		//	FileTest << iday+1 << " " << ihour << " " << hourlyFlow << endl;
		//}
		/*---debugging--------------------------------------------------------------------------*/
	}

	FileDaily.close();
	FileHourlyMax.close();
	FileHourly.close();

	/*--- debugging ------------------*/
	//FileTest.close();
	/*---------------------------------*/


	for(int i=0; i<Row; i++){
		for(int j=0; j<Col; j++){
			if(cellObject[i][j].isThisGridEnd==true || cellObject[i][j].riverNumber < 0){
				continue;
			}
			cellObject[i][j].deleteQandA(Ndays);
		}
	}

	for(int i=0; i<Row; i++){
		delete[] cellObject[i];
	}
	delete[] cellObject;

	for(int i=0; i<ROFnum_row; i++){
		delete[] rofObject[i];
	}
	delete[] rofObject;

	delete[] upperGridNum;

	return 0;
}


//	֐̒`

int makeBasinList(int row, int col, class Cell** object, int*& upperGridNum)
{
	int** basin;

	basin = new int*[row];
	for(int i=0; i<row; i++){
		
		basin[i] = new int[col];
	}


	//eIuWFNgbasinzbasinɂ
	for(int i=0; i<row; i++){
		for(int j=0; j<col; j++){
			basin[i][j] = object[i][j].basin;
		}
	}

	//ɕבւ
	int tmp;

	for(int i=0; i<row; i++){
		for(int j=0; j<col; j++){
			
			for(int jj=j+1; jj<col; jj++){
				
				if(basin[i][j] > basin[i][jj]){
					
					tmp = basin[i][j];
					basin[i][j] = basin[i][jj];
					basin[i][jj] = tmp;
				}
			}
			
			for(int ii=i+1; ii<row; ii++){
				for(int jj=0; jj<col; jj++){
					
					if(basin[i][j] > basin[ii][jj]){
						
						tmp = basin[i][j];
						basin[i][j] = basin[ii][jj];
						basin[ii][jj] = tmp;
					}
				}
			}
			
		}
	}

	//zbasin̗vfɏdĂȂ͉邩
	int num = 0;
	int dammy = 0;

	for(int i=0; i<row; i++){
		for(int j=0; j<col; j++){
			
			if(basin[i][j] > 0 && basin[i][j] != dammy){
				dammy = basin[i][j];
				num++;
			}
		}
	}

	//z upperGridNum ̃m
	upperGridNum = new int[num];

	//z upperGridNum basin̒lɓ
	int k=0;
	for(int i=0; i<row; i++){
		for(int j=0; j<col; j++){
			
			if(basin[i][j] > 0 && basin[i][j] != dammy){
				*(upperGridNum+k) = basin[i][j];
				dammy = basin[i][j];
				k++;
			}
		}
	}

	for(int i=0; i<row; i++){
		delete[] basin[i];
	}
	delete[] basin;

	return num;
}


void read2DValue(ROF** fROF, int Ndays, int ROFnum_row, int ROFnum_col)
{
	// read input data file arranged for two dimensional data. 

	char ROFfilename[256];

	//sprintf_s(ROFfilename, 256, "./input/rain.dat");
	sprintf(ROFfilename, "./input/rain.dat");
	cout << "Read " << ROFfilename << endl;

	ifstream fin(ROFfilename);
	if(!fin){
		cout << "ROF file cannot open.\n";
		exit(1);
	}

	int year, month, day, hour, xcol, yrow;
	for(int iday=1; iday <= Ndays; iday++){
		for(int ihour=0; ihour<24; ihour++){
			fin >> year >> month >> day >> hour >> xcol >> yrow;
			for(int i=ROFnum_row-1; i>=0; i--){
				for(int j=0; j<ROFnum_col; j++){
					fin >> fROF[i][j].value[iday][ihour] ;
				}
			}
		}
	}

	fin.close();
}

